// RQBSYS.PRG (alternate) - compile with option '-n'

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
This module defines an alternate RQBSYS.PRG.  Please read the
default RQBSYS.PRG first before using or modifying this module.

In this alternate mechanism, file structure definitions are
loaded from Xtrieve Data Dictionary files (.DDF).  This assumes
that when your program USEs a file, it specifies the Xtrieve
Table name, rather than the Btrieve file name (although the two
might be the same).  RaSQL will look up the table name in the
data dictionary and then load the file structure.  If a table
name is not found, RaSQL will fall back onto the 'RQB' mechanism
defined in RQBSYS.PRG.

This module does not modify the Xtrieve data dictionary files
when creating new Btrieve files.  Rather it creates an .RQB'
file.  This was done for two reasons.  First, main data tables
are usually predefined before running the application.  Second,
we felt that it was a good idea not to modify DDF files "on the
fly": these should be maintained exclusively via Xtrieve.  Of
course, feel free to enhance the code to modify the DDF files,
but use caution.

To use this module, compile it and include the OBJ in your link
script.

Special note for Clipper 5.2: RaSQL now supports index TAGs.
The tag name depends on the type of index.  If an index is
permanent, then the tag name is created in the same way as in
RQBSYS.PRG, i.e., as field names concatenated by underscore (_).
Thus, for example, if a permanent index was constructed with
fields NAME+SSNO, then the tag would be NAME_SSNO.  If the index
is supplemental, then the tag is the name of the index assigned
when the index was created in Xtrieve.  If you want all indexes
to have Xtrieve tags, then they should all be supplemental.
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define XF_ID		1		// field positions for faster access
#define XF_LOC		3
#define XE_FILE		2
#define XE_NAME		3
#define XE_DATATYPE	4
#define XE_OFFSET	5
#define XE_SIZE		6
#define XE_DEC		7

#define AFILE 241	// FILE.DDF will be open in this area
#define AFIELD 242	// FIELD.DDF will be open in this area

static XFILE := 0	// will contain AFILE if FILE.DDF is open
static XFIELD := 0	// will contain AFIELD if FIELD.DDF is open
static LASTFILE		// last filename processed by RQBFILE
static LASTNAME		// last object name processed by RQBFILE


///////////////////////////////////////////////////////////////////////////
// Called:   When file is opened via USE or DBUSEAREA.
// Syntax:   cStructure := RQBSTRUCT(cName)
// Param:    Name of passed in USE command or DBUSEAREA function.
//           Note: This should be name of Xtrieve table, not Btrieve file.
// Return:   Standard RaSQL file structure string (as used by N_XUSE or by 
//           DEFINITION clause in USE command).
// Default:  Reads information from Xtrieve DDF files
//
FUNCTION RQBSTRUCT(cName)
LOCAL fid
LOCAL iArea := N_XSELECT()
LOCAL cStructure := ""
LOCAL iOffset := 0
LOCAL iNum := 0
LOCAL cField, cNum
IF XFILE != 0 .OR. XTROPEN()
	n_xSelect(XFILE)
	n_xSeek(cName)					// seek table name
	IF !n_xEof()		
		fid := n_xFetch(XF_ID)			// table id
		n_xSelect(XFIELD)
		n_xOrder(2)				// seek by file ID
		n_xSeek(fid)				// first element
		DO WHILE n_xFetch(XE_FILE) = fid	// build string
			IF n_xFetch(XE_DATATYPE) >= 0
				IF iOffset < N_XFETCH(XE_OFFSET)
					// this part makes sure sum of 
					// preceding fields' lengths match 
					// field offsets
					cStructure += ;
						PADR("FILL"+LTRIM(STR(iOffset)),20)+;
							"C "+;
							STR(N_XFETCH(XE_OFFSET)-iOffset)+;
							".00 ;"
					iOffset := N_XFETCH(XE_OFFSET)
				ENDIF
				cField := UPPER(n_xFetch(XE_NAME))

				// this part handles field names which are
				// not unique within first 10 characters
				IF LEN(cField) >= 10 .AND. LEFT(cField,10) $ cStructure
					cNum := LTRIM(STR(++iNum))
					cField := LEFT(cField,10-LEN(cNum))+cNum
				ENDIF
				cStructure += ;
					PADR(cField,20) + ;
					PADR(N_XCVTYPE(n_xFetch(XE_DATATYPE)),2) + ;
					STR(n_xFetch(XE_SIZE),5) + "." + ;
					STR(n_xFetch(XE_DEC),2) + " ;"
			endif
			iOffset += N_XFETCH(XE_SIZE)
			n_xSkip()
		enddo
		IF UPPER(TRIM(cName)) == "X$FILE"
			cStructure := cStructure + "FILLER              C 00010.00 ;"
		ENDIF
	endif
	n_xSelect(iArea)
	XTRCLOSE()		// see notes below
ENDIF
IF cStructure == ""
	// can't find DDF entry: try using RQB file instead
	IF '.' $ cName
		cName := LEFT(cName,AT('.',cName)-1)
	ENDIF
	cStructure := fullspec(memoreadz(TRIM(cName)+".RQB"))
ENDIF
RETURN(cStructure)


///////////////////////////////////////////////////////////////////////////
// Called:   When file is opened via USE or DBUSEAREA
// Syntax:   cFile := RQBFILE(cName)
// Param:    Name passed in USE command or DBUSEAREA function.
//           Note: This should be name of Xtrieve table, not Btrieve file.
// Return:   Full Name of associated Btrieve file.
// Default:  Reads information from Xtrieve DDF files
//
function RQBFILE(cName)
local iArea := N_XSELECT()
local cFile := ""
if XFILE != 0 .or. XTROPEN()
	N_XSELECT(XFILE)
	n_xSeek(cName)					// seek table name
	if !n_xEof()
		cFile := trim(n_xFetch(XF_LOC))	// associated Btrieve file
	endif
	n_xSelect (iArea)
	XTRCLOSE()		// see notes below
endif
if cFile == ""
	// can't find DDF entry: use default extension instead
	cFile := fullspec(IIF('.'$cName,trim(cName),trim(cName)+N_XDEFEXT()))
endif
LASTFILE := cFile				// save for later
LASTNAME := cName				// save for later
return(cFile)


///////////////////////////////////////////////////////////////////////////
// Called:   When SET ORDER TO TAG is used
// Syntax:   iOrder := RQBXGET(cTag)
// Param:    Name of index Tag.  (Or names of index key fields, separated
//           by ','; no imbedded blanks).
// Returns:  Number of index
// Default:  Reads from Xtrieve DDF files
//
function RQBXGET(cName)
local fid, eid
local k, x
local i := -1
local iArea := N_XSELECT()
local cFile := N_XDBF()
local cTag := upper(trim(cName))
if XFILE != 0 .or. XTROPEN()
	n_xSelect(XFILE)
	n_xSeek(cFile)
	if !n_xEof()
		fid := n_xFetch(XF_ID)
		n_xSelect(XFIELD)
		n_xOrder(4)			// seek by file ID+field name
		n_xSeek(fid,cName)
		if !n_xEof() .and. n_xFetch(XE_DATATYPE) < 0
			i := n_xFetch(XE_OFFSET) + 1
		endif
	endif
	n_xSelect (iArea)
	XTRCLOSE()		// see notes below
endif
if i == -1
	// Can't find index tag entry
	// Use index field names to best estimate tag name
	x :=  n_xIndexkey()
	for k = 1 to x			// compare to index fields
		eid := strtran(n_xIndexkey(k),',','_')
		if cTag == eid		// exact match best
			i := k
			exit
		endif
		if cTag = eid		// else closest match
			i := k
		endif
	next k
endif
return(i)


///////////////////////////////////////////////////////////////////////////
// Called:   When ORDNAME() or similar function is called
// Syntax:   cTag := RQBXNAME(iOrder)
// Param:    Index Order
// Return:   Tag name (or index fields if no tag)
// Default:  Accesses Xtrieve DDF files
//
function RQBXNAME(iOrder)
local fid, eid
local cTag := ""
local iArea := N_XSELECT()
local cFile := N_XDBF()
if XFILE != 0 .or. XTROPEN()
	n_xSelect(XFILE)			// seek table by name
	n_xSeek(cFile)
	if !n_xEof()
		fid := n_xFetch(XF_ID)
		n_xSelect(XFIELD)
		n_xOrder(2)			// seek by file ID
		n_xSeekGT(fid)			// indexes at end of definitions
		n_xSkip(-1)
		do while n_xFetch(XE_FILE) == fid
			if n_xFetch(XE_DATATYPE) < 0 ;
				.and. n_xFetch(XE_OFFSET) + 1 == iOrder
				cTag := n_xFetch(XE_NAME)
				exit
			endif
			n_xSkip(-1)
		enddo
	endif
	n_xSelect (iArea)
	XTRCLOSE()		// see notes below
endif
if cTag == ""
	// no Tag, use index fields as tag
	cTag := strtran(n_xIndexkey(iOrder),',','_')
endif
return(cTag)


///////////////////////////////////////////////////////////////////////////
// Called:   When a Btrieve file is created
// Syntax:   RQBCREATE(cStructure)
// Param:    New file structure string
// Default:  Writes structure to RQB file based on LASTNAME
// Note   :  Bad idea to update DDFs on the fly - use Xtrieve instead
//
function RQBCREATE(cStructure)
local cFile := iif('.' $ LASTNAME , ;
		left(LASTNAME,at('.',LASTNAME)-1), ;
		trim(LASTNAME))
memowrit(cFile+".RQB",cStructure)		// create RQB file
return NIL


// just a place holder function for now
function RQBXDEL(cName)
return (0)


// just a place holder function for now
function RQBXADD(cName)
return (0)


///////////////////////////////////////////////////////////////////////////
// Called:   Internally by RQBSYS
// Syntax:   .T.|.F. := XTROPEN()
// Return:   .T. if FILE.DDX and FIELD.DDF open successfully.
// Notes:    Sets static variables XFILE and XFIELD
//
static function XTROPEN
local iArea := N_XSELECT()
local rc := .F.
local errorlvl := N_XERRLVL()
n_xSelect(AFILE)
N_XERRLVL(0)
if n_xUse("FILE.DDF",;
		"XF_ID               NI00002.00 ;" + ;
		"XF_NAME             C 00020.00 ;" + ;
		"XF_LOC              C 00064.00 ;" + ;
		"XF_FLAGS            NI00001.00 ;" + ;
		"FILLER              C 00010.00 ;")
	XFILE := AFILE
	n_xOrder(2)			// always seek by name
	n_xSelect(AFIELD)
	if n_xUse("FIELD.DDF", ;
			"XE_ID               NI00002.00 ;" + ;
			"XE_FILE             NI00002.00 ;" + ;
			"XE_NAME             C 00020.00 ;" + ;
			"XE_DATATYPE         NI00001.00 ;" + ;
			"XE_OFFSET           NI00002.00 ;" + ;
			"XE_SIZE             NI00002.00 ;" + ;
			"XE_DEC              NI00001.00 ;" + ;
			"XE_FLAGS            NI00002.00 ;")
		XFIELD := AFIELD
		rc := .T.
	endif
endif
if !rc
	if XFIELD > 0
		n_xSelect(XFIELD)
		n_xClose()
	endif
	XFILE := XFIELD := 0
endif
n_xSelect (iArea)
N_XERRLVL(errorlvl)
return rc


///////////////////////////////////////////////////////////////////////////
// Called:   Internally by RQBSYS
// Syntax:   XTRCLOSE()
// Return:   nil
// Notes:    Closes X$FILE and X$FIELD
//           Comment the code in this routine for slightly faster
//           performance.  However, if you do so you may not be able to
//           open files FILE.DDF or FIELD.DDF in your program.  You 
//           should rarely need to do this, though
//
static function XTRCLOSE
local iArea := N_XSELECT()
IF XFILE != 0
	n_xSelect(XFILE)
	n_xClose()
	n_xSelect(XFIELD)
	n_xClose()
	XFILE := XFIELD := 0
	n_xSelect(iArea)
ENDIF
RETURN nil


////////////////////////////////////////////////////////////////////////
// Called:   Internally by RQBSYS
// Syntax:   cPath := fullspec(cFile)
// Param:    file name
// Return:   Full path based on N_XPATH or SET PATH
//
STATIC FUNCTION fullspec(cFname)
LOCAL cSpec := ""
LOCAL cPath := N_XPATH()
IF !EMPTY(cPath) .AND. ;
		(.NOT. ( ':' $ cFname .OR. '\' = cFname .OR. '/' = cFname)) ;
		.AND. FILE(cPath + cFname)	// look in N_XPATH
	RETURN (cPath + cFname)
ENDIF
CH_FILE(cFname,@cSpec)
RETURN cSpec


////////////////////////////////////////////////////////////////////////
// Called:   Internally by RQBSYS
// Syntax:   cString := memoreadz(cFile)
// Param:    Text file name
// Return:   Contents of text file with trailing eofs (hex 1A) 
//           and nulls (hex 00) removed
//
static function memoreadz(fname)
local cStructure := memoread(fname)
local k := at(chr(26),cStructure)
if k > 0
	cStructure := left(cStructure,k-1)
endif
k := at(chr(0),cStructure)
if k > 0
	cStructure := left(cStructure,k-1)
endif
return cStructure

////////////////////////////////////////////////////////////////////////
// IMPORTANT: include this code in any modification/copy of RQBSYS.PRG
//
announce RQBRDD
init procedure RQBSYS()
	extern _VRQBRDD
	rddRegister("RQBRDD",1)
	return

////////////////////////////////////////////////////////////////////////
// RECOMMENDED: include this code in any modification/copy of RQBSYS.PRG
//
exit procedure RQBSYSEXIT()
	N_XLOGOUT()		// ensures Btrieve cleanup
	return

// eos: RQBSYS.PRG (alternate)
