// RQBSYS.PRG - compile with option '-n' 
#include "SET.CH"

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
This module, RQBSYS.PRG, allows you to customize the way RaSQL
obtains file header information for Btrieve files.

Btrieve files, unlike DBFs, do not contain file headers defining
the fields in the record.  Traditionally, it has been up to the
program to read the Btrieve record as a single block and parse it
into its component memory variables.  This has a major
disadvantage, however, of requiring the program to change
whenever a file's structure changes.  RaSQL gets around this
problem by defining a "pseudo" file header for each Btrieve file.
The "pseudo" header is a string, conforming to the format
described in the N_XUSE function, stored in a file with the same
name as the Btrieve file, but with a file type of ".RQB".  For
example, CLIENT.DTA would have its header information stored in
CLIENT.RQB.  (However, so would CLIENT.DAT; so it is important
that you use distinct file names for each Btrieve file.)
Whenever your program USEs the Btrieve file, RaSQL will look for
the associated RQB file and load the file structure (assuming you
have not specified a DEFINITION clause in the USE command).

Since the RQB methodology is completely arbitrary on our part
(although it is simple and it works) we have decided to open it
up so that the technique could be modified.  For example, you
might have your own data dictionary scheme and wish to load file
definitions from that; or you might wish to load your file
definitions from Xtrieve.  Modifying or replacing the file
RQBSYS.PRG will enable you to customize RaSQL's operation to your
own needs.  (In fact, the file ALT_RQB.PRG contains alternate
RQBSYS code that loads definitions from Xtrieve databases.  To
use it, compile ALT_RQB and include it in your link script.)

If you decide to use RaSQL's default "RQB" mechanism, no special
action is needed on your part, since this standard RQBSYS.PRG is
included in the RaSQL library.  However, if you wish to modify
the default operation, then you should copy this file to a new
file (say, MYRQBSYS.PRG), modify it, compile it, and include the
OBJ file in your link script, much like you would define an
alternate ERRORSYS.PRG or GETSYS.PRG.

We strongly suggest that you have a good idea of what you are
doing before you play around with it, though.

Special note for Clipper 5.2: RaSQL now supports index TAGs.
However, the standard RQBSYS has no built in mechanism for
handling tag names.  Thus, tag names are constructed as the field
names included in the index, concatenated by underscore (_).  For
example, if an index is defined as UPPER(CITY)+STR(SSNO), then
the tag would be CITY_SSNO.
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

memvar MBM_STRUCTURE, MBM_NAME

static LASTFILE		// last filename processed by RQBFILE
static LASTNAME		// last object name processed by RQBFILE


////////////////////////////////////////////////////////////////////////
// Called:   When file is opened via USE or DBUSEAREA.
// Syntax:   cStructure := RQBSTRUCT(cName)
// Param:    Filename passed to USE command or DBUSEAREA function
// Return:   Standard RaSQL file structure string (as used by N_XUSE or
//           by DEFINITION clause in USE command).
// Default:  Reads information from RQB file or MBM file if one or the 
//           other exists.
//
function RQBSTRUCT(cName)
LOCAL cFile := iif('.' $ cName , left(cName,at('.',cName)-1),trim(cName))
LOCAL cSpec := fullspec(cFile+".RQB")
LOCAL cStructure := NIL
IF !EMPTY(cSpec)
	cStructure := memoreadz(cSpec)
ENDIF
IF EMPTY(cStructure)		// RQB not found
	cSpec := fullspec(cFile+".MBM")
	IF !EMPTY(cSpec)			// see if old-style MBM
		restore from (cSpec) additive
		cStructure := MBM_STRUCTURE
		release MBM_STRUCTURE, MBM_NAME
	ENDIF
ENDIF
RETURN(cStructure)


////////////////////////////////////////////////////////////////////////
// Called:   When file is opened via USE or DBUSEAREA
// Syntax:   cFile := RQBFILE(cName)
// Param:    Name passed in USE command or DBUSEAREA function
// Return:   Full Name of Btrieve file with extension
// Default:  Adds default extension to filename if none has been specified.
//
function RQBFILE(cName)
local cFile := fullspec(iif( '.' $ cName, cName, trim(cName)+N_XDEFEXT()))
LASTFILE := cFile
LASTNAME := cName
return (cFile)


////////////////////////////////////////////////////////////////////////
// Called:   When a Btrieve file is created
// Syntax:   RQBCREATE(cStructure)
// Param:    New file structure string
// Default:  Writes structure to RQB file based on LASTNAME
//
function RQBCREATE(cStructure)
local cFile := iif('.' $ LASTNAME , ;
		left(LASTNAME,at('.',LASTNAME)-1), ;
		trim(LASTNAME))
memowrit(cFile+".RQB",cStructure)		// create RQB file
return NIL


////////////////////////////////////////////////////////////////////////
// Called:   When SET ORDER TO TAG is used
// Syntax:   iOrder := RQBXGET(cTag)
// Param:    Name of index Tag
// Returns:  Number of index
// Default:  Use index field names to best estimate tag name
//
function RQBXGET(cName)
local k
local key
local i := -1
local cTag := upper(trim(cName))
local x :=  n_xIndexkey()
if empty(cName)
	return 0
endif
for k = 1 to x			// compare to index fields
	key := strtran(n_xIndexkey(k),',','_')
	if cTag == key		// exact match best
		i := k
		exit
	endif
	if cTag = key		// else closest match
		i := k
	endif
next k
return(i)


////////////////////////////////////////////////////////////////////////
// Called:   When INDEX ON ... TAG is used
// Note:     Not supported with standard RQB files.
//
function RQBXADD(cName)
return -1


////////////////////////////////////////////////////////////////////////
// Called:   When DELETE TAG is used
// Note:     Not supported with standard RQB files.
//
function RQBXDEL(cName)
return (-1)


////////////////////////////////////////////////////////////////////////
// Called:   When ORDNAME() or similar function is called
// Syntax:   cTag := RQBXNAME(iOrder)
// Param:    Index Order
// Return:   "Pseudo" tag name made up of index field names separated 
//           by underscore (_)
//
FUNCTION RQBXNAME(iOrder)
local cTag := strtran(n_xIndexkey(iOrder),',','_')
return cTag


////////////////////////////////////////////////////////////////////////
// Called:   Internally by RQBSYS
// Syntax:   cString := memoreadz(cFile)
// Param:    Text file name
// Return:   Contents of text file with trailing eofs (hex 1A) 
//           and nulls (hex 00) removed
//
static function memoreadz(fname)
local cStructure := memoread(fname)
local k := at(chr(26),cStructure)
if k > 0
	cStructure := left(cStructure,k-1)
endif
k := at(chr(0),cStructure)
if k > 0
	cStructure := left(cStructure,k-1)
endif
return cStructure

////////////////////////////////////////////////////////////////////////
// Called:   Internally by RQBSYS
// Syntax:   cPath := fullspec(cFile)
// Param:    file name
// Return:   Full path based on N_XPATH or SET PATH
//
STATIC FUNCTION fullspec(cFname)
LOCAL cSpec := ""
LOCAL cPath := N_XPATH()
IF !EMPTY(cPath) .AND. ;
		(.NOT. ( ':' $ cFname .OR. '\' = cFname .OR. '/' = cFname)) ;
		.AND. FILE(cPath + cFname)	// look in N_XPATH
	RETURN (cPath + cFname)
ENDIF
CH_FILE(cFname,@cSpec)
RETURN cSpec


////////////////////////////////////////////////////////////////////////
// IMPORTANT: include this code in any modification/copy of RQBSYS.PRG
//
announce RQBRDD
init procedure RQBSYS()
	extern _VRQBRDD
	rddRegister("RQBRDD",1)
	return

////////////////////////////////////////////////////////////////////////
// RECOMMENDED: include this code in any modification/copy of RQBSYS.PRG
//
exit procedure RQBSYSEXIT()
	N_XLOGOUT()		// ensures Btrieve cleanup
	return

// eos: RQBSYS.PRG
